/*
  works together with the deviceConfigServer <-> producer Interface
  Copyright (C) 1998  Martin Vogt

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation.

  For more information look at the file COPYRIGHT in this package

 */



#include <devices/genericStreamProducer.h>



static int producerCounter=0;

static void *writerThread(void *arg){
  ((GenericStreamProducer*)arg)->writeloop();
  return NULL;
}   




GenericStreamProducer::GenericStreamProducer() {
  server=new DeviceConfigServer();
  server->setNotifyClient(this);
  config=server->getDeviceConfig();
  
  lExit=false;
  lServerEnter=false;
  instanz=producerCounter;
  producerCounter++;
  lastFileName=new Buffer(40);

  pthread_create(&tr,NULL,writerThread,this);

}


GenericStreamProducer::~GenericStreamProducer() {
  void* ret;

  writeOutLock();
  lExit=true;
  writeOutUnlock();
  pthread_join(tr,&ret);

  delete server;
  delete lastFileName;
}



int GenericStreamProducer::open(char* filename ) {
  if (producer != NULL) {
    cout <<"GenericStreamProducer"<<filename<<endl;
    producer->open(filename);
    lastFileName->clear();
    lastFileName->append(filename);
  }
  return true;
}

int GenericStreamProducer::close() {
  if (producer != NULL) {
    producer->close();
  }
  return true;
}


int GenericStreamProducer::play() {
  if (producer != NULL) {
    // this implements a common interface behaviour.
    // if we are stopped and we click "play"
    // we play the last file we have already played
    if (getState() == _STATUS_STOPPED) {
      cout << "lastFile:"<<lastFileName->getData()<<endl;
      producer->open(lastFileName->getData());
    } else {
      producer->play();
    }
  }
  return true;
}


int GenericStreamProducer::pause() {
  if (producer != NULL) {

    if (getState() == _STATUS_STOPPED) {
      return true;
    }
    // This is a common interface behaviour.
    // If we are paused and we hit "pause" again
    // we try a "play"
    if (getState() == _STATUS_PAUSED) {
      producer->play();
    } else {
      producer->pause();
    }
  }
  return true;
}


int GenericStreamProducer::jump(int sec) {
  if (producer != NULL) {
    producer->jump(sec);
  }
  return true;
}


int GenericStreamProducer::getState() {
  int back;
  StatusInfo* statusInfo=config->getStatusInfo();
  back=statusInfo->getStatus();
  return back; 
}


void GenericStreamProducer::setProducer(Producer* producer) {
  writeOutLock();
  this->producer=producer;
  server->setProducer(producer);
  writeOutUnlock();
}


Producer* GenericStreamProducer::getProducer() {
  return producer;
}


//
// ThreadState Interface [START]
//


void GenericStreamProducer::setProducerEnter(int lServerEnter) {
  writeOutLock();
  this->lServerEnter=lServerEnter;
  writeOutUnlock();
}



void GenericStreamProducer::enqueueThreadState(int state) {
  cout << "enqueueThreadState vor lock"<<endl;
  writeOutLock();
  cout << "enqueue state: -s"<<state<<endl;
  ThreadState::enqueueThreadState(state);
  cout << "enqueue state: -e"<<state<<endl;
  writeOutUnlock();
  cout << "enqueueThreadState nach lock"<<endl;
}



//
// ThreadState Interface [END]
//


/**
   The writeloop method is a single thread. It must be sure that
   the thread never blocks in the function below this.
   This is the only method where the thread can sleep.

   The thread delivers two different informations.
   One information is a statusInformation. If it delivers
   this no other info in the deviceConfig class is new.
   (no new audio pcm sample, no new musictitle etc,...)
   Only the status _change_ is delivered.

   If there is no status change to deliver, the thread
   delivers the things with minor priority. eg. pcm samples,...

*/

void GenericStreamProducer::writeloop() {
  DeviceConfigArray* configArray=new DeviceConfigArray();
  AudioBuffer* audioBuffer=config->getAudioBuffer();
  MemChunk* memChunk;
  MemChunk* emptyChunk=new MemChunk(0);

  config->setStreamProducer(this);
  pthread_mutex_lock(&deliverMut);

  while(lExit==false) {
    if (pthread_mutex_trylock(&changeMut) == EBUSY) {
      cout << "writeloop locked because of change request!"
	   << "instanz:"<<instanz<<endl;
      pthread_cond_wait(&deliverCond,&deliverMut);
      cout << "writeloop continues change request ready"
	   << "instanz:"<<instanz<<endl;
      continue;
    }
    pthread_mutex_unlock(&changeMut);
    pthread_mutex_unlock(&deliverMut);
    waitForUnblock(); 
    pthread_mutex_lock(&deliverMut);
    

    if ( (getDeliverMode() == _DELIVER_MODE_SINGLE) &&
	 (getData() == true) ) {
      cout << "delivermode single && data sleeping"<<endl;
      pthread_cond_wait(&deliverCond,&deliverMut);
      cout << "delivermode single && data sleeping --wakeup"<<endl;
      continue;
    }
    
    // Now two possibilities:
    // 1. we have _no_ thread state to deliver (then we deliver data)
    // 2. wa have a thread state. (then we deliver it)
    
    if (hasState() == true) {
      StatusInfo* statusInfo=config->getStatusInfo();
      AudioBuffer* audioBuffer=config->getAudioBuffer();

      statusInfo->setStatus(dequeueThreadState());
      statusInfo->setStatusChange(true);
      cout << "dequeueThreadState:"<<statusInfo->getStatus()<<endl;
      audioBuffer->setMemChunk(emptyChunk);
    } else {
      if (lServerEnter== false) {
	cout << "lServerEnter== false (sleeping)"
	     << "instanz:"<<instanz<<endl;
	pthread_cond_wait(&deliverCond,&deliverMut);
	cout << "lServerEnter==  wakeup"
	     << "instanz:"<<instanz<<endl;
	continue;
      }       
      server->updateDeviceConfig(config);
      StatusInfo* statusInfo=config->getStatusInfo();
      statusInfo->setStatusChange(false);
    }
    

    // ok now we have read data. Make sure this is
    // delivered to the device.
    if ( hasListener() ) {
      configArray->clear();
      configArray->setDeviceConfigAt(0,config);
      configArray->setEntries(1);
      config->setStreamProducer(this);
      memChunk=audioBuffer->getMemChunk();
      memChunk->addLock();
      writeOut(configArray);
      memChunk->removeLock();
      dataSendNotify();
    } else {
      cout << " instanz:"<<instanz<<"kein listener sleeping"<<endl;
      pthread_cond_wait(&deliverCond,&deliverMut);
      cout << " instanz:"<<instanz<<" wakeup sleeping"<<endl;
    }
  }
  pthread_mutex_unlock(&deliverMut);
  delete configArray;
  delete emptyChunk;
}
  
 

